# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Copyright 2012 Thomi Richards

from gi.repository import GObject, Gtk, GLib
import logging

from server_monitor import get_server_monitor
from settings import get_settings

logger = logging.getLogger(__name__)

class JenkinsServerModel(GObject.GObject):
    """A model for a jenkins server."""

    name = GObject.property(type=str)
    url = GObject.property(type=str)
    refresh_period = GObject.property(type=int, minimum=30, maximum=60*60, default=5*60)

    __gsignals__ = {
        'job_added':        (GObject.SIGNAL_RUN_LAST,
                            GObject.TYPE_NONE, # signal return type
                            (GObject.TYPE_PYOBJECT,)), # signal parameters
        'job_removed':      (GObject.SIGNAL_RUN_LAST,
                            GObject.TYPE_NONE,
                            (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, name, server_settings):
        GObject.GObject.__init__(self)
        self.name = name
        self.url = server_settings['url']
        self.refresh_period = server_settings['refresh_period']

        self._server_monitor = get_server_monitor()
        self._jobs = {}
        self.enabled = False

        self.connect("notify::url", self._on_url_changed)
        logger.info("Server model for '%s' created with enabled=%r", self.name, self.enabled)
        if self.url != "":
            self._enable_polling()

    def _start_get_server_data(self):
        if self.enabled:
            logger.debug("Starting job refresh")
            self._server_monitor.start_get_server_data(self.url, self._on_server_data_arrived)
        return self.enabled

    def _on_server_data_arrived(self, data):
        status,data = data
        logger.debug("Download finished.")
        if status:
            self._update_jobs(data)
        else:
            logger.error("Error while trying to get data from server: %r", data)
            pass # TODO - emit error signal so the user sees this failure.

    def _update_jobs(self, jobs_list):
        """Update the internal jobs dictionary with details form this data retrieval."""
        # add and update:
        monitored_jobs = get_settings().get_server_by_name(self.name)['monitor_jobs']
        for job in jobs_list:
            job_name = job['name']
            if not self._jobs.has_key(job_name):
                job_obj = JenkinsJobModel(
                            job_name,
                            job['url'],
                            job['color'],
                            job_name in monitored_jobs,
                            self)

                self._jobs[job_name] = job_obj
                self.emit('job-added', job_obj)
                logger.debug("Found new job: %s", job_name)
            else:
                if self._jobs[job_name].color != job['color']:
                    self._jobs[job_name].set_job_color(job['color'])
        to_remove  =[]
        names = [j['name'] for j in jobs_list]
        for name, job in self._jobs.iteritems():
            if name not in names:
                # Jobs is not on this server. Maybe the server has changed, or
                # maybe the job has been removed, but we need to remove it now:
                to_remove.append(name)
        for job_name in to_remove:
            self._remove_job(job_name)

    def get_jobs(self):
        """Get a list of job objects in this server, sorted alphabetically by
        job name.

        """
        job_names = self._jobs.keys()
        job_names.sort()
        return [self._jobs[name] for name in job_names]

    def get_job_by_name(self, job_name):
        """Get a JenkinsJobModel instance by name, or None if it does not exist
        on this server.

        """
        return self._jobs.get(job_name, None)

    def _on_url_changed(self, sender, prop):
        if not self.enabled and self.url != "":
            logger.info("Server model '%s' url changed to '%s'. Starting data refresh.", self.name, self.url)
            self._enable_polling()
        elif self.enabled and self.url == "":
            logger.info("Server model %s url changed to ''.", self.name)
            self.enabled = False
            self._clear_all_jobs()
        if self.enabled:
            #url has changed, so we should do an explicit retrieve straight away:
            self._clear_all_jobs()
            self._start_get_server_data()

    def _clear_all_jobs(self):
        names = self._jobs.keys()
        for n in names:
            self._remove_job(n)

    def _remove_job(self, job_name):
        logger.debug("Removing job '%s' as it is no longer on the server.", job_name)
        self._jobs[job_name].monitored = False
        self.emit('job-removed', self._jobs[job_name])
        self._jobs.pop(job_name)

    def _enable_polling(self):
        logger.debug("Enableding polling for server monitor.")
        if not self.enabled:
            self.enabled = True
            GLib.timeout_add(self.refresh_period * 1000, self._start_get_server_data)
            self._start_get_server_data()



class JenkinsJobModel(GObject.GObject):
    """A model for a jenkins server job."""

    name = GObject.property(type=str)
    url = GObject.property(type=str)
    build = GObject.property(type=int)
    color = GObject.property(type=str)
    description = GObject.property(type=str)
    health_description = GObject.property(type=str)
    health_score = GObject.property(type=int, minimum=0, maximum=100)
    monitored = GObject.property(type=bool, default=False)

    def __init__(self, name, url, color, monitored, server):
        GObject.GObject.__init__(self)
        self.name = name
        self.url = url
        self.set_job_color(color)
        self.monitored = monitored
        self._server_model = server
        self._server_monitor = get_server_monitor()
        self.connect('notify::monitored', self._on_monitored_changed)
        if self.monitored:
           self._enable_polling()

    def start_get_more_details(self):
        """Start downloading more information from the jenkins server."""
        if self.monitored:
            self._server_monitor.start_get_job_data(self._server_model.url, self.name, self._on_job_data_arrived)
        return self.monitored

    def _on_job_data_arrived(self, data):
        state,job_info = data
        if state:
            self.set_job_color(job_info['color'])
            if self.description != job_info['description']:
                self.description = job_info['description']
            score, desc = self._get_health_details(job_info['healthReport'])
            if self.health_description != desc:
                self.health_description = desc
            if self.health_score != score:
                self.health_score = score
            buildnum = job_info['lastBuild']['number']
            if self.build != buildnum:
                self.build = buildnum
        else:
            logger.error("Error while trying to get job information from server: %r", data)

    def _get_health_details(self, health_report):
        """Jenkins server sends us a health report, which is a list of health
        indicators.

        Be pessimistic and return the score and description of the worst one.
        """
        score = None
        default_description = "No health indicators present."
        description = []

        for health_indication in health_report:
            if score is None or health_indication['score'] < score:
                score = health_indication['score']
            description.append(health_indication['description'])
        if score is None:
            score = 0

        return score, '\n'.join(description) if description else default_description

    def set_job_color(self, color):
        """Sets the job color, with some translation."""
        if color == 'blue':
            self.set_job_color('green')
        elif color == 'aborted':
            self.set_job_color('grey')
        else:
            if self.color != color:
                self.color = color

    def _enable_polling(self):
        logger.debug("Starting server data polling for job '%s'", self.name)
        GLib.timeout_add(self._server_model.refresh_period * 1000, self.start_get_more_details)
        self.start_get_more_details()

    def _on_monitored_changed(self, sender, prop):
        if self.monitored:
            self._enable_polling()

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self.name)

GObject.type_register(JenkinsServerModel)
GObject.type_register(JenkinsJobModel)
