# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Copyright 2012 Thomi Richards

from gi.repository import GdkPixbuf
from mock import patch
import os.path
import shutil
import tempfile
from testtools import TestCase
from testtools.matchers import Contains, IsInstance, raises, Equals, Not, Raises

import graphics


def write_valid_svg_to_path(path):
    with open(path, 'w') as f:
        f.write(r'''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg
   xmlns="http://www.w3.org/2000/svg"
   xmlns:xlink="http://www.w3.org/1999/xlink"
   version="1.0"
   width="500"
   height="500"
   id="Blank image">
  <rect
     style="fill:#FFFFFF"
     x="0"
     y="0"
     width="500"
     height="500"
     id="White section" />
</svg>
''')


class GraphicsCacheTests(TestCase):
    """Tests for the graphics cache store."""

    def test_make_key(self):
        # We really don't need to test this, but it's an easy test to write, so
        # why not?
        key = graphics._cache_make_key("foobar", 123)
        self.assertThat(key, Equals("foobar123"))

    @patch.dict(graphics._pixbuf_cache, dict(f24=42))
    def test_cache_has_image_at_size(self):
        """cache_has_key must return False when the dict is empty."""
        self.assertThat(graphics.cache_has_image_at_size('f', 24), Equals(True))
        self.assertThat(graphics.cache_has_image_at_size('f', 28), Equals(False))
        self.assertThat(graphics.cache_has_image_at_size('z', 12), Equals(False))

    @patch.dict(graphics._pixbuf_cache)
    def test_cache_get_image_throws_keyerror(self):
        """cache_get_image_at_size must raise KeyError if the key does not exist."""
        fn = lambda: graphics.cache_get_image_at_size('foobar', 24)
        self.assertThat(fn, raises(KeyError))

    def test_cache_get_image_works(self):
        """cache_get_image must return whatever is in the cache."""
        class Foo:
            pass
        with patch.dict(graphics._pixbuf_cache, dict(f24=Foo())):
            result = graphics.cache_get_image_at_size('f', 24)
            self.assertThat(result, IsInstance(Foo))

    @patch.dict(graphics._pixbuf_cache)
    def test_cache_set_image_throws(self):
        """cache_set_image must throw TypeError if we try and store something
        other than a Pixbuf instance.

        """
        for obj in (42, 0.3, True, "Hi", u'there', self,):
            fn = lambda: graphics.cache_set_image_at_size("foo", 24, obj)
            self.assertThat(fn, raises(TypeError))

    @patch.dict(graphics._pixbuf_cache)
    def test_cache_set_image_doesnt_throw(self):
        """cache_set_image must not throw when we give it a valid pixbuf."""
        pb = GdkPixbuf.Pixbuf()
        fn = lambda: graphics.cache_set_image_at_size("foo", 24, pb)
        self.assertThat(fn, Not(Raises()))

    @patch.dict(graphics._pixbuf_cache)
    def test_cache_set_image_works(self):
        """cache_set_image must store image in dictionary."""
        pb = GdkPixbuf.Pixbuf()
        graphics.cache_set_image_at_size("f", 24, pb)
        self.assertThat(graphics._pixbuf_cache, Contains("f24"))
        self.assertThat(graphics._pixbuf_cache["f24"], Equals(pb))


class SystemGraphicStoreTests(TestCase):
    """Tests for the system graphics store."""

    def setUp(self):
        super(SystemGraphicStoreTests, self).setUp()
        if not graphics._icon_loader:
            self.skipTest("IconTheme class not loaded (perhaps X11 is not present?)")

    @patch('graphics._icon_loader', spec=True)
    def test_system_has_image(self, mock_obj):
        """system_has_image must defer to the icon loader."""
        mock_obj.has_icon.return_value=False
        self.assertThat(graphics.system_has_image("foobar"), Equals(False))
        mock_obj.has_icon.assert_called_once_with("foobar")

    @patch('graphics._icon_loader', spec=True)
    def test_system_get_image_raises_IOError(self, mock_obj):
        """system_get_image must raise IOError if the icon isn't available."""
        mock_obj.has_icon.return_value=False
        fn = lambda: graphics.system_get_image("foobar", 24)
        self.assertThat(fn, raises(IOError))

    @patch('graphics._icon_loader', spec=True)
    def test_system_get_image_works(self, mock_obj):
        """system_get_image must return icon loaded from icon loader."""
        mock_obj.has_icon.return_value = True
        mock_obj.load_icon.return_value = 42
        result = graphics.system_get_image("foobar", 24)
        self.assertThat(result, Equals(42))
        mock_obj.load_icon.assert_called_once_with("foobar", 24, 0)


class LocalStoreGraphicsTests(TestCase):
    """Tests for the local graphic store."""

    def setUp(self):
        super(LocalStoreGraphicsTests, self).setUp()
        self.tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tempdir)

    def make_svg_in_local_store(self, filename, readable=True):
        path = os.path.join(self.tempdir, filename)
        write_valid_svg_to_path(path)
        if not readable:
            os.chmod(path, 0)

    def test_local_store_get_folder_value(self):
        """Test the default value for the local storage location."""
        expected = os.path.join(
            os.path.dirname(graphics.__file__),
            'images')
        actual = graphics.local_store_get_folder()
        self.assertThat(expected, Equals(actual))

    @patch('graphics.local_store_get_folder', spec=True)
    def test_has_image_returns_false_when_not_exist(self, mock_get_path):
        """local_store_has_image must return false if the file does not exist."""
        mock_get_path.return_value = self.tempdir
        result = graphics.local_store_has_image('foobar')
        self.assertThat(result, Equals(False))

    @patch('graphics.local_store_get_folder', spec=True)
    def test_has_image_returns_false_when_now_readable(self, mock_get_path):
        """local_store_has_image must return false if the file exists but is not readable."""
        mock_get_path.return_value = self.tempdir
        self.make_svg_in_local_store('foobar.svg', readable=False)
        result = graphics.local_store_has_image('foobar')
        self.assertThat(result, Equals(False))

    @patch('graphics.local_store_get_folder', spec=True)
    def test_has_image_returns_true(self, mock_get_path):
        """local_store_has_image must return true if the file exists and is readable."""
        mock_get_path.return_value = self.tempdir
        self.make_svg_in_local_store('foobar.svg')
        result = graphics.local_store_has_image('foobar')
        self.assertThat(result, Equals(True))

    @patch('graphics.local_store_get_folder', spec=True)
    def test_get_image_throws_ioerror(self, mock_get_path):
        """local_store_get_image must throw ioerror if the image is missing."""
        mock_get_path.return_value = self.tempdir
        fn = lambda: graphics.local_store_get_image('foobar', 24)
        self.assertThat(fn, raises(IOError))

    @patch('graphics.local_store_get_folder', spec=True)
    def test_get_image_works(self, mock_get_path):
        mock_get_path.return_value = self.tempdir
        self.make_svg_in_local_store('foobar.svg')
        result = graphics.local_store_get_image('foobar', 24)

        self.assertThat(result, IsInstance(GdkPixbuf.Pixbuf))
        self.assertThat(result.get_width(), Equals(24))
        self.assertThat(result.get_height(), Equals(24))


@patch.dict(graphics._pixbuf_cache)
class LoadPixbufByNameTests(TestCase):
    """Tests for the load_bixbuf_by_name function."""

    @patch('graphics.cache_has_image_at_size', lambda x, y: False)
    @patch('graphics.system_has_image', lambda x: False)
    @patch('graphics.local_store_has_image', lambda x: False)
    def test_throws_runtime_error(self):
        """If none of the stores have the image, must raise a RuntimeError."""
        fn = lambda: graphics.load_pixbuf_by_name('foobar', 24)
        self.assertThat(fn, raises(RuntimeError))

    @patch('graphics.cache_has_image_at_size', spec=True)
    @patch('graphics.cache_get_image_at_size', spec=True)
    @patch('graphics.system_has_image', spec=True)
    @patch('graphics.local_store_has_image', spec=True)
    def test_prefers_cache_over_all(self, mock_local_has, mock_sys_has, mock_cache_get, mock_cache_has):
        """Must prefer the cache over all other sources."""
        mock_cache_has.return_value=True
        mock_cache_get.return_value=42
        mock_sys_has.return_value=False
        mock_local_has.return_value=False

        result = graphics.load_pixbuf_by_name('foobar', 24)

        self.assertThat(result, Equals(42))
        mock_cache_has.assert_called_once_with('foobar', 24)
        mock_cache_get.assert_called_once_with('foobar', 24)
        self.assertThat(mock_sys_has.called, Equals(False))
        self.assertThat(mock_local_has.called, Equals(False))

    @patch('graphics.cache_has_image_at_size', spec=True)
    @patch('graphics.system_get_image', spec=True)
    @patch('graphics.system_has_image', spec=True)
    @patch('graphics.local_store_has_image', spec=True)
    @patch('graphics.cache_set_image_at_size', lambda x, y, z: None)
    def test_prefers_sys_over_local(self, mock_local_has, mock_sys_has, mock_sys_get, mock_cache_has):
        """Must prefer the system store over local store."""
        mock_cache_has.return_value=False
        mock_sys_has.return_value=True
        mock_sys_get.return_value=42
        mock_local_has.return_value=False

        result = graphics.load_pixbuf_by_name('foobar', 24)

        self.assertThat(result, Equals(42))
        mock_cache_has.assert_called_once_with('foobar', 24)
        mock_sys_has.assert_called_once_with('foobar')
        mock_sys_get.assert_called_once_with('foobar', 24)
        self.assertThat(mock_local_has.called, Equals(False))

    @patch('graphics.cache_has_image_at_size', spec=True)
    @patch('graphics.local_store_get_image', spec=True)
    @patch('graphics.system_has_image', spec=True)
    @patch('graphics.local_store_has_image', spec=True)
    @patch('graphics.cache_set_image_at_size', lambda x, y, z: None)
    def test_uses_local_last(self, mock_local_has, mock_sys_has, mock_local_get, mock_cache_has):
        """Must prefer the system store over local store."""
        mock_cache_has.return_value=False
        mock_sys_has.return_value=False
        mock_local_has.return_value=True
        mock_local_get.return_value=42

        result = graphics.load_pixbuf_by_name('foobar', 24)

        self.assertThat(result, Equals(42))
        mock_cache_has.assert_called_once_with('foobar', 24)
        mock_sys_has.assert_called_once_with('foobar')
        mock_local_has.assert_called_once_with('foobar')
        mock_local_get.assert_called_once_with('foobar', 24)

    @patch('graphics.cache_has_image_at_size', lambda x, y: False)
    @patch('graphics.local_store_has_image', lambda x: True)
    @patch('graphics.local_store_get_image', lambda x, y: 42)
    @patch('graphics.cache_set_image_at_size', spec=True)
    def test_sets_cache_from_system(self, mock_cache_set):
        """Must set the returned image in the cache when loaded from system store."""
        graphics.load_pixbuf_by_name('foobar', 24)
        mock_cache_set.assert_called_once_with('foobar', 24, 42)


class GetPixbufForHealthScoreTests(TestCase):
    """Tests for the get_pixbuf_for_health_score function."""

    def setUp(self):
        super(GetPixbufForHealthScoreTests, self).setUp()
        self.call_count = 0;
        self.expected_args = ()
        self.expected_kwargs = {}

    def test_throws_typeerror(self):
        """Function must throw typeerror when passed something other than an int for the score."""
        for obj in (0.1, True, 'boo', u'bang', self):
            fn = lambda: graphics.get_pixbuf_for_health_score(obj, 24)
            self.assertThat(fn, raises(TypeError))

    def test_throws_valueerror(self):
        """Function must throw valueError if score is less than 0 or greather than 100."""
        should_throw = (
            lambda: graphics.get_pixbuf_for_health_score(-1, 24),
            lambda: graphics.get_pixbuf_for_health_score(101, 24)
            )
        shouldnt_throw = (lambda: graphics.get_pixbuf_for_health_score(s, 24) for s in range(101))
        for fn in should_throw:
            self.assertThat(fn, raises(ValueError))
        for fn in shouldnt_throw:
            self.assertThat(fn, Not(Raises()))

    def test_0_to_19(self):
        """Function must load the correct icon for values less than 20."""
        self.expected_args = ("health-00to19", 24)
        with patch('graphics.load_pixbuf_by_name', self.call_counter):
            for score in range(20):
                graphics.get_pixbuf_for_health_score(score, 24)
        self.assertThat(self.call_count, Equals(20))

    def test_20_to_39(self):
        """Function must load the correct icon for values less than 20."""
        self.expected_args = ("health-20to39", 24)
        with patch('graphics.load_pixbuf_by_name', self.call_counter):
            for score in range(20, 40):
                graphics.get_pixbuf_for_health_score(score, 24)
        self.assertThat(self.call_count, Equals(20))

    def test_40_to_59(self):
        """Function must load the correct icon for values less than 20."""
        self.expected_args = ("health-40to59", 24)
        with patch('graphics.load_pixbuf_by_name', self.call_counter):
            for score in range(40, 60):
                graphics.get_pixbuf_for_health_score(score, 24)
        self.assertThat(self.call_count, Equals(20))

    def test_60_to_79(self):
        """Function must load the correct icon for values less than 20."""
        self.expected_args = ("health-60to79", 24)
        with patch('graphics.load_pixbuf_by_name', self.call_counter):
            for score in range(60, 80):
                graphics.get_pixbuf_for_health_score(score, 24)
        self.assertThat(self.call_count, Equals(20))

    def test_80_to_100(self):
        """Function must load the correct icon for values less than 20."""
        self.expected_args = ("health-80plus", 24)
        with patch('graphics.load_pixbuf_by_name', self.call_counter):
            for score in range(80, 101):
                graphics.get_pixbuf_for_health_score(score, 24)
        self.assertThat(self.call_count, Equals(21))

    def call_counter(self, *args, **kwargs):
        """Utility function. Counts the number of times called, and asserts arguments are correct."""
        self.call_count += 1
        self.assertThat(args, Equals(self.expected_args))
        self.assertThat(kwargs, Equals(self.expected_kwargs))


class GetPixbufForStateTests(TestCase):
    """Tests for the get_pixbuf_for_state function."""

    @patch('graphics.load_pixbuf_by_name', spec=True)
    def test_translates_blue_to_green(self, mock_load):
        """Function must translate requests for 'blue' to 'green'."""
        graphics.get_pixbuf_for_state('blue', 24)
        mock_load.assert_called_once_with('status-green', 24)

    @patch('graphics.load_pixbuf_by_name', spec=True)
    def test_loads_green(self, mock_load):
        """Function must load the green state correctly."""
        graphics.get_pixbuf_for_state('green', 34)
        mock_load.assert_called_once_with('status-green', 34)

    @patch('graphics.load_pixbuf_by_name', spec=True)
    def test_loads_yellow(self, mock_load):
        """Function must load the yellow state correctly."""
        graphics.get_pixbuf_for_state('yellow', 45)
        mock_load.assert_called_once_with('status-yellow', 45)

    @patch('graphics.load_pixbuf_by_name', spec=True)
    def test_loads_red(self, mock_load):
        """Function must load the red state correctly."""
        graphics.get_pixbuf_for_state('red', 56)
        mock_load.assert_called_once_with('status-red', 56)

    @patch('graphics.load_pixbuf_by_name', spec=True)
    def test_loads_grey(self, mock_load):
        """Function must load the grey state correctly."""
        graphics.get_pixbuf_for_state('grey', 67)
        mock_load.assert_called_once_with('status-grey', 67)

    def test_translates_unknown_to_grey(self):
        """Function must translate all unknown names to grey."""
        for state_name in ('disabled', 'unknown', 'wibble', 'bang'):
            with patch('graphics.load_pixbuf_by_name', spec=True) as p:
                graphics.get_pixbuf_for_state(state_name, 12)
                p.assert_called_once_with('status-grey', 12)


class CombineTwoImagesTests(TestCase):
    """Tests for the combine_two_images function."""

    def setUp(self):
        super(CombineTwoImagesTests, self).setUp()
        self.tempdir = tempfile.mkdtemp()
        self.svg_file_path = os.path.join(self.tempdir, 'test.svg')
        write_valid_svg_to_path(self.svg_file_path)
        self.addCleanup(shutil.rmtree, self.tempdir)

    def get_pixbuf(self, width, height):
        pb = GdkPixbuf.Pixbuf.new_from_file_at_scale(self.svg_file_path, width, height, False)
        self.assertThat(pb.get_height(), Equals(height))
        self.assertThat(pb.get_width(), Equals(width))
        return pb

    def test_throws_typeerror(self):
        """Must throw typerror if either the left or right images are not Pixbuf instances."""
        pb = self.get_pixbuf(50, 50)
        invalid_types = (1, .2, True, False, self, "Foo", u'sdf')
        for obj in invalid_types:
            fn = lambda: graphics.combine_two_images(obj, pb)
            self.assertThat(fn, raises(TypeError))
        for obj in invalid_types:
            fn = lambda: graphics.combine_two_images(pb, obj)
            self.assertThat(fn, raises(TypeError))

    def test_throws_valueerror_on_different_heights(self):
        """Must throw ValueError if the image heights are different."""
        pb1 = self.get_pixbuf(50, 50)
        pb2 = self.get_pixbuf(50, 100)
        fn = lambda: graphics.combine_two_images(pb1, pb2)
        self.assertThat(fn, raises(ValueError))

    def test_throws_valueerror_on_different_widths(self):
        """Must throw ValueError if the image heights are different."""
        pb1 = self.get_pixbuf(50, 50)
        pb2 = self.get_pixbuf(100, 50)
        fn = lambda: graphics.combine_two_images(pb1, pb2)
        self.assertThat(fn, raises(ValueError))

    def test_result_has_correct_dimensions(self):
        """Result must be twice as wide as original images, and same height."""
        pb1 = self.get_pixbuf(100, 100)
        pb2 = self.get_pixbuf(100, 100)
        result = graphics.combine_two_images(pb1, pb2)
        self.assertThat(result, IsInstance(GdkPixbuf.Pixbuf))
        self.assertThat(result.get_height(), Equals(100))
        self.assertThat(result.get_width(), Equals(200))

@patch.dict(graphics._pixbuf_cache)
class BasicGraphicsTests(TestCase):
    """Test that we have the art assets we need."""

    GUARANTEED_IMAGE_NAMES = (
        'health-00to19',
        'health-20to39',
        'health-40to59',
        'health-60to79',
        'health-80plus',
        'jenkins',
        'status-blue',
        'status-green',
        'status-grey',
        'status-red',
        'status-yellow',
        )

    def test_assets_exist(self):
        """Test that all the art assets we need are found.."""
        for img_name in self.GUARANTEED_IMAGE_NAMES:
            pb = graphics.load_pixbuf_by_name(img_name, 24)
            self.assertThat(pb, IsInstance(GdkPixbuf.Pixbuf))

    def test_can_load_different_sizes(self):
        """We musn't cache different size images."""
        sizes = (8,16,24,100,500)
        for size in sizes:
            for img_name in self.GUARANTEED_IMAGE_NAMES:
                pb = graphics.load_pixbuf_by_name(img_name, size)
                self.assertThat(pb.get_width(), Equals(size))
                self.assertThat(pb.get_height(), Equals(size))

    def test_caching_works(self):
        img_name = self.GUARANTEED_IMAGE_NAMES[0]
        pb1 = graphics.load_pixbuf_by_name(img_name, 24)
        pb2 = graphics.load_pixbuf_by_name(img_name, 24)
        self.assertThat(id(pb1), Equals(id(pb2)))
        self.assertThat(pb1.get_pixels(), Equals(pb2.get_pixels()))
